@@ -107,6 +107,10 @@ function (formula, tau = 0.5, data, subset, weights, na.action, method = "br",
     if(method == "model.frame")return(mf)
     mt <- attr(mf, "terms")
     weights <- as.vector(model.weights(mf))
+    tau <- sort(unique(tau))
+    eps <- .Machine$double.eps^(2/3)
+    if (any(tau == 0)) tau[tau == 0] <- eps
+    if (any(tau == 1)) tau[tau == 1] <- 1 - eps
     Y <- model.response(mf)
     if(method == "sfn"){
 	if(requireNamespace("MatrixModels") && requireNamespace("Matrix")){
@@ -116,18 +120,18 @@ function (formula, tau = 0.5, data, subset, weights, na.action, method = "br",
 	    mf$x <- X
 	    }
     }	
-    else
+    else{
 	X <- model.matrix(mt, mf, contrasts)
-    eps <- .Machine$double.eps^(2/3)
+	vnames <- dimnames(X)[[2]]
+    }
     Rho <- function(u,tau) u * (tau - (u < 0))
-    if(length(tau)>1){
-	if(any(tau < 0) || any(tau > 1))
-		stop("invalid tau:  taus should be >= 0 and <= 1")
-	if(any(tau == 0)) tau[tau == 0] <- eps
-	if(any(tau == 1)) tau[tau == 1] <- 1 -eps
-	coef <- matrix(0,ncol(X),length(tau))
-        rho <- rep(0,length(tau))
-	fitted <- resid <- matrix(0,nrow(X),length(tau))
+    if (length(tau) > 1) {
+      if (any(tau < 0) || any(tau > 1)) 
+        stop("invalid tau:  taus should be >= 0 and <= 1")
+      coef <- matrix(0, ncol(X), length(tau))
+      rho <- rep(0, length(tau))
+      if(!(method %in% c("ppro","qfnb","pfnb"))){
+        fitted <- resid <- matrix(0, nrow(X), length(tau))
 	for(i in 1:length(tau)){
 	    z <- {if (length(weights))
                  	rq.wfit(X, Y, tau = tau[i], weights, method, ...)
@@ -139,8 +143,8 @@ function (formula, tau = 0.5, data, subset, weights, na.action, method = "br",
 	    fitted[,i] <- Y - z$residuals
 	   }
 	taulabs <- paste("tau=",format(round(tau,3)))
-	dimnames(coef) <- list(dimnames(X)[[2]],taulabs)
-	dimnames(resid) <- list(dimnames(X)[[1]],taulabs)
+	dimnames(coef) <- list(vnames, taulabs)
+	dimnames(resid)[[2]] <- taulabs
 	fit <- z
         fit$coefficients <-  coef
         fit$residuals <- resid
@@ -149,22 +153,35 @@ function (formula, tau = 0.5, data, subset, weights, na.action, method = "br",
 	else if(method == "scad") class(fit) <- c("scadrqs","rqs")
 	else class(fit) <- "rqs"
 	}
+      else if(method == "pfnb"){ # Preprocessing in fortran loop
+	  fit <- rq.fit.pfnb(X, Y, tau)
+          class(fit) = "rqs" 
+          }
+      else if(method == "qfnb"){ # simple fortran loop method 
+	  fit <- rq.fit.qfnb(X, Y, tau)
+          class(fit) = ifelse(length(tau) == 1,"rq","rqs") 
+          }
+      else if(method == "ppro"){ # Preprocessing method in R
+	  fit <- rq.fit.ppro(X, Y, tau, ...)
+          class(fit) = ifelse(length(tau) == 1,"rq","rqs") 
+          }
+    }
     else{
-        process <- (tau < 0 || tau > 1)
-	if(tau == 0) tau <- eps
-	if(tau == 1) tau <- 1 -eps
-        fit <- {
-            if (length(weights))
-                rq.wfit(X, Y, tau = tau, weights, method, ...)
-            else rq.fit(X, Y, tau = tau, method, ...)
-           }
-	if(process)
-		rho <- list(x = fit$sol[1,],y = fit$sol[3,])
-	else {
-		if(length(dim(fit$residuals)))
-		    dimnames(fit$residuals) <- list(dimnames(X)[[1]],NULL)
-                rho <-  sum(Rho(fit$residuals,tau))
-		}
+      process <- (tau < 0 || tau > 1)
+      if(process && method != "br") 
+        stop("when tau not in [0,1] method br must be used")
+      fit <- {
+        if(length(weights))
+          rq.wfit(X, Y, tau = tau, weights, method, ...)
+        else rq.fit(X, Y, tau = tau, method, ...)
+        }
+      if(process)
+	rho <- list(tau = fit$sol[1,], rho = fit$sol[3,])
+      else {
+	if(length(dim(fit$residuals)))
+          dimnames(fit$residuals) <- list(dimnames(X)[[1]],NULL)
+          rho <-  sum(Rho(fit$residuals,tau))
+          } 
 	if(method == "lasso") class(fit) <- c("lassorq","rq")
 	else if(method == "scad") class(fit) <- c("scadrq","rq")
         else class(fit) <- ifelse(process, "rq.process", "rq")
@@ -188,17 +205,20 @@ function (formula, tau = 0.5, data, subset, weights, na.action, method = "br",
 "rq.fit" <-
 function(x, y, tau = 0.5, method = "br", ...)
 {
-    if(length(tau) > 1) {
-	    tau <- tau[1]
-	    warning("Multiple taus not allowed in rq.fit: solution restricted to first element")
-	}
+    if(length(tau) > 1 && method != "ppro") {
+	tau <- tau[1]
+        warning("Multiple taus not allowed in rq.fit: solution restricted to first element")
+        }
 
     fit <- switch(method,
 		fn = rq.fit.fnb(x, y, tau = tau, ...),
 		fnb = rq.fit.fnb(x, y, tau = tau, ...),
 		fnc = rq.fit.fnc(x, y, tau = tau, ...),
 		sfn = rq.fit.sfn(x, y, tau = tau, ...),
+		conquer = rq.fit.conquer(x, y, tau = tau, ...),
 		pfn = rq.fit.pfn(x, y, tau = tau, ...),
+		pfnb = rq.fit.pfnb(x, y, tau = tau, ...),
+		ppro= rq.fit.ppro(x, y, tau = tau, ...),
 		br = rq.fit.br(x, y, tau = tau, ...),
 		lasso = rq.fit.lasso(x, y, tau = tau, ...),
 		scad = rq.fit.scad(x, y, tau = tau, ...),
@@ -419,9 +439,9 @@ function (object, newdata, type = "Qhat", stepfun = FALSE, na.action = na.pass,
        else if(type == "Fhat"){
 	  taus <- c(taus[1],taus)
           if(M > 1) 
-	      f <- apply(pred,2,function(y) stepfun(y,t))
+	      f <- apply(pred,2,function(y) stepfun(y,taus))
           else
-              f <- stepfun(pred[,1],c(taus[1],taus))
+              f <- stepfun(pred[,1],taus)
           }
        else stop("Stepfuns must be either 'Qhat' or 'Fhat'")
        return(f)
@@ -438,7 +458,7 @@ function (object, newdata, type = "Qhat", stepfun = FALSE, na.action = na.pass,
         if(M > 1) 
 	    f <- apply(pred[-1,], 2, function(z) akjfun(z, p, ...))
         else
-	   akjfun(pred[,1], p, ...)
+	   f = akjfun(pred[,1], p, ...)
        return(f)
        }
     else return(t(pred))
@@ -644,6 +664,19 @@ function (x, y, tau = 0.5, alpha = 0.1, ci = FALSE, iid = TRUE,
     }
 }
 
+"rq.fit.conquer" <- function(x, y, tau = 0.5, kernel = c("Gaussian", "uniform", 
+    "parabolic", "triangular"), h = 0,  tol = 1e-04, 
+    iteMax = 5000, ci = FALSE, alpha = 0.05, B = 200)
+{
+    if(!requireNamespace("conquer", quietly = TRUE))
+            stop("method conquer requires package conquer")
+    fit = conquer::conquer(x[,-1], y, tau = tau, kernel = kernel, h = h, 
+	tol = tol, iteMax = iteMax, ci = FALSE, alpha = alpha, B = 1000)
+    coefficients = fit$coeff 
+    names(coefficients) = dimnames(x)[[2]]
+    residuals = fit$residual
+    list(coefficients = coefficients, tau = tau, residuals = residuals)
+}
 "rq.fit.fnb" <-
 function (x, y, tau = 0.5, rhs = (1-tau)*apply(x,2,sum), beta = 0.99995, eps = 1e-06)
 {
@@ -661,13 +694,13 @@ function (x, y, tau = 0.5, rhs = (1-tau)*apply(x,2,sum), beta = 0.99995, eps = 1
         c = as.double(-y), rhs = as.double(rhs), d = as.double(d),as.double(u),
         beta = as.double(beta), eps = as.double(eps),
         wn = as.double(wn), wp = double((p + 3) * p),
-        it.count = integer(3), info = integer(1))
+        nit = integer(3), info = integer(1))
     if (z$info != 0)
-        stop(paste("Error info = ", z$info, "in stepy: singular design"))
+        warning(paste("Error info = ", z$info, "in stepy: possibly singular design"))
     coefficients <- -z$wp[1:p]
     names(coefficients) <- dimnames(x)[[2]]
     residuals <- y - x %*% coefficients
-    list(coefficients=coefficients, tau=tau, residuals=residuals)
+    list(coefficients=coefficients, tau=tau, residuals=residuals, nit = z$nit)
 }
 
 "rq.fit.fnc" <-
@@ -767,7 +800,7 @@ function (x, y, tau = 0.5, alpha = 3.2, lambda = 1, start = "rq", beta = 0.9995,
 
 
 "rq.fit.lasso" <-
-function (x, y, tau = 0.5, lambda = 1, beta = 0.9995, eps = 1e-06)
+function (x, y, tau = 0.5, lambda = 1, beta = 0.99995, eps = 1e-06)
 {
     n <- length(y)
     p <- ncol(x)
@@ -791,7 +824,7 @@ function (x, y, tau = 0.5, lambda = 1, beta = 0.9995, eps = 1e-06)
     d <- rep(1, N)
     u <- rep(1, N)
     wn <- rep(0, 10 * N)
-    wn[1:N] <- rep(0.5, N)
+    wn[1:N] <- 0.5
     z <- .Fortran("rqfnb", as.integer(N), as.integer(p), a = as.double(t(as.matrix(X))),
         c = as.double(-Y), rhs = as.double(rhs), d = as.double(d),
         as.double(u), beta = as.double(beta), eps = as.double(eps),
@@ -821,7 +854,7 @@ function (x, y, tau = 0.5, lambda = 1, beta = 0.9995, eps = 1e-06)
 # is already done in the rq.fit.fnb calls.
 #
 function(x, y, tau = 0.5,  Mm.factor = 0.8,
-	max.bad.fixup = 3, eps = 1e-6)
+	max.bad.fixups = 3, eps = 1e-6)
 {
 	#rq function for n large --
 	n <- length(y)
@@ -830,13 +863,16 @@ function(x, y, tau = 0.5,  Mm.factor = 0.8,
 	if(tau < 0 | tau > 1)
 		stop("tau outside (0,1)")
 	p <- ncol(x)
-	m <- round(((p + 1) * n)^(2/3))
+	m <- round(sqrt(p) * n^(2/3))
 	not.optimal <- TRUE
+	ifix = 0
+	ibad = 0
 	while(not.optimal) {
+		ibad = ibad + 1
 		if(m < n)
 			s <- sample(n, m)
 		else {
-			b <- rq.fit.fnb(x, y, tau = tau,  eps = eps)$coef
+			z <- rq.fit.fnb(x, y, tau = tau,  eps = eps)
 			break
 		}
 		xx <- x[s,  ]
@@ -852,8 +888,9 @@ function(x, y, tau = 0.5,  Mm.factor = 0.8,
 		kappa <- quantile(r/pmax(eps, band), c(lo.q, hi.q))
 		sl <- r < band * kappa[1]
 		su <- r > band * kappa[2]
-		bad.fixup <- 0
-		while(not.optimal & (bad.fixup < max.bad.fixup)) {
+		bad.fixups <- 0
+		while(not.optimal & (bad.fixups < max.bad.fixups)) {
+			ifix = ifix + 1
 			xx <- x[!su & !sl,  ]
 			yy <- y[!su & !sl]
 			if(any(sl)) {
@@ -876,24 +913,23 @@ function(x, y, tau = 0.5,  Mm.factor = 0.8,
 			su.bad <- (r < 0) & su
 			sl.bad <- (r > 0) & sl
 			if(any(c(su.bad, sl.bad))) {
-				if(sum(su.bad | sl.bad) > 0.10000000000000001 *
-					M) {
-					warning("Too many fixups:  doubling m")
-					m <- 2 * m
-					break
+				if(sum(su.bad | sl.bad) > 0.1 * M) {
+				    # This warning may get annoying?
+				    warning("Too many fixups:  doubling m")
+				    bad.fixups <- bad.fixups + 1
+				    m <- 2 * m
+				    break
 				}
 				su <- su & !su.bad
 				sl <- sl & !sl.bad
-				bad.fixup <- bad.fixup + 1
 			}
 			else not.optimal <- FALSE
 		}
 	}
-	coefficients <- b
+	nit <- c(z$nit,ifix,ibad)
+	coefficients <- z$coef
 	names(coefficients) <- dimnames(x)[[2]]
-	residuals <- y - x %*% b
-	return(list(coefficients=coefficients, tau=tau,
-		residuals=residuals))
+	list(coefficients=coefficients, tau=tau, nit = nit)
 }
 
 "rq.wfit" <-
@@ -914,7 +950,10 @@ function(x, y, tau = 0.5, weights, method = "br",  ...)
 		br = rq.fit.br(wx, wy, tau = tau, ...),
 		fnc = rq.fit.fnc(wx, wy, tau = tau, ...),
 		sfn = rq.fit.sfn(wx, wy, tau = tau, ...),
-                pfn = rq.fit.pfn(wx, wy, tau = tau, ...), {
+		conquer = rq.fit.conquer(wx, wy, tau = tau, ...),
+		ppro = rq.fit.ppro(wx, wy, tau = tau, ...),
+                pfn = rq.fit.pfn(wx, wy, tau = tau, ...), 
+                pfnb = rq.fit.pfnb(wx, wy, tau = tau, ...), {
 			what <- paste("rq.fit.", method, sep = "")
 			if(exists(what, mode = "function"))
 				(get(what, mode = "function"))(x, y, ...)
@@ -981,7 +1020,7 @@ function (object, ...) {
         attr(val,"n") <- n
 	if(pen){
 	   if(!hasArg(edfThresh)) edfThresh <- 0.0001
-           attr(val,"df") <- sum(abs(object$coefficients) > edfThresh)
+           attr(val,"df") <- apply(abs(object$coefficients) > edfThresh,2,sum)
 	  }
 	else  attr(val,"df") <- p
         class(val) <- "logLik"
@@ -1045,32 +1084,40 @@ c(edf, aic)
 #		"xy"	uses xy-pair method
 #		"wxy"	uses weighted (generalized) method
 #		"pwy"	uses the parzen-wei-ying method
+#		"pxy"	uses the preprocessing method
 #		"mcmb"	uses the Markov chain marginal bootstrap method
 #		"cluster"  uses the Hagemann clustered wild gradient method
-#		"BLB"	uses the Bag of Little Boostraps method
+#		"conquer"  uses the He et al multiplier bootstrap
+#		"BLB"	uses the Bag of Little Bootstraps method
 #
 #
 function (object, se = NULL, covariance = FALSE, hs = TRUE, U = NULL, gamma = 0.7, ...)
 {
     if(object$method == "lasso")
          stop("no inference for lasso'd rq fitting: try rqss (if brave, or credulous)")
+    if(object$method == "conquer") se = "conquer"
     mt <- terms(object)
     m <- model.frame(object)
     y <- model.response(m)
     dots <- list(...)
-    if(object$method == "sfn")
+    method <- object$method
+    if(object$method == "sfn"){
 	x <- object$model$x
-    else
+	vnames <- names(object$coef)
+	ctrl <- object$control
+    }
+    else{
 	x <- model.matrix(mt, m, contrasts = object$contrasts)
+	vnames <- dimnames(x)[[2]]
+    }
     wt <- as.vector(model.weights(object$model))
     tau <- object$tau
     eps <- .Machine$double.eps^(1/2)
     coef <- coefficients(object)
     if (is.matrix(coef))
         coef <- coef[, 1]
-    vnames <- dimnames(x)[[2]]
     resid <- object$residuals
-    n <- length(resid)
+    n <- length(y)
     p <- length(coef)
     rdf <- n - p
     if (!is.null(wt)) {
@@ -1103,17 +1150,25 @@ function (object, se = NULL, covariance = FALSE, hs = TRUE, U = NULL, gamma = 0.
     else if (se == "nid") {
         h <- bandwidth.rq(tau, n, hs = hs)
 	while((tau - h < 0) || (tau + h > 1)) h <- h/2
-        bhi <- rq.fit.fnb(x, y, tau = tau + h)$coef
-        blo <- rq.fit.fnb(x, y, tau = tau - h)$coef
+        bhi <- rq.fit(x, y, tau = tau + h, method = method)$coef
+        blo <- rq.fit(x, y, tau = tau - h, method = method)$coef
         dyhat <- x %*% (bhi - blo)
         if (any(dyhat <= 0))
             warning(paste(sum(dyhat <= 0), "non-positive fis"))
         f <- pmax(0, (2 * h)/(dyhat - eps))
         fxxinv <- diag(p)
-        fxxinv <- backsolve(qr(sqrt(f) * x)$qr[1:p, 1:p,drop=FALSE], fxxinv)
-        fxxinv <- fxxinv %*% t(fxxinv)
-        cov <- tau * (1 - tau) * fxxinv %*% crossprod(x) %*%
-            fxxinv
+	if(method == "sfn"){
+	    D <- t(x) %*% (f * x)
+            D <- chol(0.5 * (D + t(D)), nsubmax = ctrl$nsubmax,
+                      nnzlmax = ctrl$nnzlmax, tmpmax = ctrl$tmpmax)
+            fxxinv <- backsolve(D, fxxinv)
+        }
+	else{
+	    fxxinv <- backsolve(qr(sqrt(f) * x)$qr[1:p, 1:p,drop=FALSE], fxxinv)
+	    fxxinv <- fxxinv %*% t(fxxinv)
+	}
+	xx <- t(x) %*% x
+        cov <- tau * (1 - tau) * fxxinv %*% xx %*% fxxinv
         scale <- mean(f)
         serr <- sqrt(diag(cov))
     }
@@ -1139,13 +1194,15 @@ function (object, se = NULL, covariance = FALSE, hs = TRUE, U = NULL, gamma = 0.
 	         cluster <- dots$cluster[-object$na.action]
 	         bargs <- modifyList(bargs, list(cluster = cluster))
 	    }
+	    if(class(bargs$x)[1] == "matrix.csr")
+	         bargs <- modifyList(bargs, list(control = ctrl))
 	B <- do.call(boot.rq, bargs)
 	}
 	else
-	    B <- boot.rq(x, y, tau, ...)
+	    B <- boot.rq(x, y, tau, coef = coef, ...)
         cov <- cov(B$B)
         serr <- sqrt(diag(cov))
-    }
+        }
    else if (se == "BLB"){ # Bag of Little Bootstraps
         n <- length(y)
         b <- ceiling(n^gamma)
@@ -1160,10 +1217,67 @@ function (object, se = NULL, covariance = FALSE, hs = TRUE, U = NULL, gamma = 0.
 	cov <- cov(B$B)
         serr <- apply(Z, 1, mean)
     }
+   else if(se == "extreme"){
+    tau0 <- tau
+    if(tau > 0.5) {
+	y <- -y
+	tau <- 1 - tau
+    }
+    if(length(dots$mofn)) mofn = dots$mofn
+    else mofn = floor(n/5)
+    if(length(dots$mofn)) kex = dots$kex
+    else kex = 20
+    if(length(dots$alpha)) alpha = dots$alpha
+    else alpha = 0.1
+    if(length(dots$R)) R = dots$R
+    else R = 200
+    m <- (tau * n + kex)/(tau * n)
+    taub <- min(tau * n/mofn, tau + (.5 - tau)/3)
+    # This warning is a bit different than Section 1.3.4 of the Handbook chapter
+    #if (tau.b.e == tau.e + (.5 - tau.e) / 3 && b >= min(n / 3, 1000))
+    #	    warning("tau may be non-extremal results are not likely to differ from central inference");
+    xbar <- apply(x,2, mean)
+    b0 <- rq.fit(x, y, tau, method = method)$coef
+    bm <- rq.fit(x, y, tau = m*tau, method = method)$coef
+    An <- (m-1)*tau * sqrt(n/(tau*(1-tau)))/c(crossprod(xbar,bm - b0))
+    bt <- rq.fit(x, y, tau=taub, method = method)$coef
+    s <- matrix(sample(1:n, mofn * R, replace = T), mofn, R)
+    mbe <- (taub * mofn + kex)/(taub * mofn)
+    bmbeb <- rq.fit(x, y, tau = mbe*taub, method = method)$coef
+    # Accelerated xy bootstrap
+    B0 <- boot.rq.pxy(x, y, s, taub, bt, method = method)
+    Bm <- boot.rq.pxy(x, y, s, tau = mbe * taub, bmbeb, method = method)
+    B <- (mbe - 1) * taub * sqrt(mofn/(taub * (1-taub))) *
+	(B0 - b0)/c((Bm - B0) %*% xbar) 
+    if (tau0 <= 0.5) {
+        bbc <- b0 - apply(B, 2, quantile, .5, na.rm = TRUE)/An
+        ciL <- b0 - apply(B, 2, quantile, 1 - alpha/2, na.rm = TRUE)/An
+        ciU <- b0 - apply(B, 2, quantile, alpha/2, na.rm = TRUE)/An
+    } else {
+        bbc <- -(b0 - apply(B, 2, quantile, .5, na.rm = TRUE)/An)
+        ciL <- -(b0 - apply(B, 2, quantile, alpha/2, na.rm = TRUE)/An)
+        ciU <- -(b0 - apply(B, 2, quantile, 1 - alpha/2, na.rm = TRUE)/An)
+    }
+    B <- R-sum(is.na(B[,1]))
+    coef <- cbind(b0, bbc, ciL, ciU)
+    if(tau0 > 0.5) {coef <-  -coef; tau <- tau0}
+    dimnames(coef) = list(dimnames(x)[[2]], c("coef", "BCcoef","ciL", "ciU"))
+}
+   else if(se == "conquer"){
+       if(length(dots$R)) R = dots$R
+       else R = 200
+       Z <- conquer(x[,-1], y, tau, ci = TRUE, B = R)
+       #Fixme:  should have option to choose another bsmethod
+       coef <- cbind(Z$coef, Z$perCI)
+       cnames <- c("coefficients", "lower bd", "upper bd")
+       dimnames(coef) <- list(vnames, cnames)
+       resid <- y - x %*% Z$coef
+   }
+
     if( se == "rank"){
 	coef <- f$coef
 	}
-    else {
+    else if(!(se %in% c("conquer", "extreme"))){
     	coef <- array(coef, c(p, 4))
     	dimnames(coef) <- list(vnames, c("Value", "Std. Error", "t value",
              "Pr(>|t|)"))
@@ -1304,3 +1418,160 @@ function (x, y, taus = c(.1,.3,.5), weights = c(.7,.2,.1),
     if(any(is.na(coefficients)))stop("NA coefs:  infeasible problem?")
     list(coefficients = coefficients, nit = z$it.count, flag = z$info)
 }
+#preprocessing for the QR process
+"rq.fit.ppro" <- function (x, y, tau, weights=NULL, Mm.factor = 0.8, eps = 1e-06, ...) 
+{
+  ntau <- length(tau)
+  n <- length(y)
+  if (nrow(x) != n) stop("x and y don't match n")
+  p <- ncol(x)
+  m <- n * sqrt(p) * max(diff(tau)) # check this length(tau) == 1 case?
+  dots <- list(...)
+  method <- ifelse(length(dots$pmethod), dots$pmethod, "fn")
+  if(length(weights)){
+      if(any(weights < 0))
+	 stop("negative weights not allowed")
+      if(length(weights) != n)
+	 stop("weights not of length(y)")
+      else {
+         x <- x * weights
+         y <- y * weights
+      }
+  }
+  coef <- matrix(NA, p, ntau)
+  resid <- matrix(NA, n, ntau)
+  rho <- rep(0, ntau)
+  Rho <- function(u, tau) u * (tau - (u < 0))
+  z <- rq.fit(x, y, tau=tau[1], method = method)
+  r <- z$resid
+  coef[,1] <- z$coef
+  rho[1] <- sum(Rho(z$residuals, tau[1]))
+  xxinv <- solve(chol(crossprod(x)))
+  # Is pmax really necessary here?
+  band <- pmax(eps, sqrt(((x %*% xxinv)^2) %*% rep(1, p)))
+  for(i in 2:ntau){
+    not.optimal <- TRUE
+    mm <- m
+    while (not.optimal) {
+      M <- Mm.factor * mm
+      lo.q <- max(1/n, tau[i] - M/(2 * n))
+      hi.q <- min(tau[i] + M/(2 * n), (n - 1)/n)
+      kappa <- quantile(r/band, c(lo.q, hi.q))
+      sl <- r < band * kappa[1]
+      su <- r > band * kappa[2]
+      while (not.optimal) {
+        xx <- x[!su & !sl, ]
+        yy <- y[!su & !sl]
+        if (any(sl)) {
+          glob.x <- c(t(x[sl, , drop = FALSE]) %*% rep(1, sum(sl)))
+          glob.y <- sum(y[sl])
+          xx <- rbind(xx, glob.x)
+          yy <- c(yy, glob.y)
+        }
+        if (any(su)) {
+          ghib.x <- c(t(x[su, , drop = FALSE]) %*% rep(1, sum(su)))
+          ghib.y <- sum(y[su])
+          xx <- rbind(xx, ghib.x)
+          yy <- c(yy, ghib.y)
+        }
+        z <- rq.fit(xx, yy, tau = tau[i], method = method)
+        b <- z$coef
+        r <- y - x %*% b
+        su.bad <- (r < 0) & su
+        sl.bad <- (r > 0) & sl
+        bad.signs <- sum(su.bad | sl.bad)
+        if (bad.signs > 0) {
+          if (bad.signs > 0.1 * M) {
+            mm <- 2*mm
+            warning("Too many fixups:  doubling m")
+            break
+          }
+          su <- su & !su.bad
+          sl <- sl & !sl.bad
+        }
+        else not.optimal <- FALSE
+      }
+    }
+    coef[,i] <- b
+    resid[,i] <- y - x %*% b
+    rho[i] <- sum(Rho(resid, tau[i]))
+  }
+  dimnames(coef) <- list(dimnames(x)[[2]], tau)
+  list(coefficients=coef, residuals = resid, rho=rho, weights = weights)
+}
+
+# R function for fnb call for multiple taus
+rq.fit.qfnb <- function(x,y,tau){
+    n <- nrow(x)
+    p <- ncol(x)
+    m <- length(tau)
+    d <- rep(1, n)
+    u <- rep(1, n)
+    z <- .Fortran("qfnb",
+		  n = as.integer(n),
+		  p = as.integer(p),
+		  m = as.integer(m),
+		  a = as.double(t(x)),
+		  y = as.double(-y),
+		  t = as.double(tau),
+		  r = double(p),
+		  d = as.double(d),
+		  u = as.double(u),
+		  wn = double(n*9),
+		  wp = double(p*(p+3)),
+		  B = double(p*m),
+		  nit = integer(3),
+		  info = integer(1))
+    if(z$info != 0)
+	warning(paste("Info = ", z$info, "in stepy: singular design: nit = ", z$nit[1]))
+    coefficients <- matrix(-z$B, p, m)
+    dimnames(coefficients) <- list(dimnames(x)[[2]],paste("tau = ",tau))
+    list(coefficients = coefficients, nit = z$nit, flag = z$info)
+}
+rq.fit.pfnb <- function (x, y, tau, m0 = NULL, eps = 1e-06) {
+    m <- length(tau)
+    n <- length(y)
+    if(!is.matrix(x)) dim(x) <- c(n,1) 
+    if (nrow(x) != n) 
+        stop("x and y don't match n")
+    p <- ncol(x)
+    if(!length(m0))
+	m0 <- floor(n^(2/3) * sqrt(p)) # Needs testing!
+    s <- sample(n,m0)
+    xs <- x[s,,drop = FALSE]
+    ys <- y[s]
+    z <- rq.fit(xs, ys, tau = tau[1], method = "fn")
+    r <- y - x %*% z$coef
+    b <- matrix(0,p,m)
+    nit <- matrix(0,5,m)
+    xxinv <- solve(chol(crossprod(xs)))
+    band <- pmax(eps, sqrt(((x %*% xxinv)^2) %*% rep(1, p)))
+    z <- .Fortran("pfnb",
+		  as.integer(n),
+		  as.integer(p),
+		  as.integer(m),
+		  as.double(t(x)),
+		  as.double(y),
+		  as.double(tau),
+		  as.double(r),
+		  b = as.double(-b),
+		  as.double(band),
+		  as.integer(m0),
+		  double(n),
+		  double(n),
+		  double(n*9),
+		  double(p*(p+3)),
+		  double(p*n),
+		  double(n),
+		  integer(n),
+		  integer(n),
+		  double(p),
+		  double(p),
+		  double(p),
+		  nit = as.integer(nit),
+		  info = integer(m))
+    coefficients <- matrix(-z$b,p,m)
+    nit <- matrix(z$nit,5,m)
+    dimnames(coefficients) <- list(dimnames(x)[[2]],paste("tau = ",tau))
+    list(coefficients = coefficients, nit = nit, flag = z$info)
+}
