@@ -32,7 +32,7 @@ function (tt, special, order = 1)
 
 "qss" <-
 function (x, constraint = "N", lambda = 1, ndum = 0, dummies = NULL, 
-    w = rep(1, length(x))) 
+    Dorder = 1, w = rep(1, length(x))) 
 {
     if (is.matrix(x)) {
         if (ncol(x) == 2) 
@@ -42,9 +42,20 @@ function (x, constraint = "N", lambda = 1, ndum = 0, dummies = NULL,
             x <- as.vector(x)
         else stop("qss objects must have dimension 1 or 2")
     }
-    if (is.vector(x)) 
-        qss <- qss1(x, constraint = constraint, lambda = lambda, 
-            dummies = dummies, ndum = ndum, w = w)
+    if (is.vector(x)) {
+	if(Dorder == 1)
+	    qss <- qss1(x, constraint = constraint, lambda = lambda, 
+		dummies = dummies, ndum = ndum, w = w)
+	else if(Dorder == 0){
+	    if(!(constraint %in% c("N","I","D")))
+		stop("This constraint not implemented for Dorder = 0")
+	    else
+		qss <- qts1(x, constraint = constraint, lambda = lambda, 
+		    dummies = dummies, ndum = ndum, w = w)
+	}
+	else
+	    stop("Dorder must be either 0 or 1")
+    }
     qss
 }
 
@@ -112,6 +123,38 @@ function(x, y, constraint = "N", lambda = 1, ndum= 0, dummies = NULL, w=rep(1,le
         lambda = lambda, A = A[, -1], R = R[, -1], r = r)
 }
 
+qts1 <- function (x, constraint = "N", lambda = 1, dummies = dummies, 
+    ndum = 0, w = rep(1, length(x))) 
+{ # This is the taut string TV(g) not TV(g') penalty option
+    xun <- unique(x[order(x)])
+    h <- diff(xun)
+    nh <- length(h)
+    nx <- length(x)
+    p <- nh + 1
+    makeD <- function(p) {
+        new("matrix.csr", ra = c(rbind(rep(-1, (p - 1)), rep(1,
+            (p - 1)))), ja = as.integer(c(rbind(1:(p - 1), 2:p))),
+            ia = as.integer(2 * (1:p) - 1), dimension = as.integer(c(p -
+                1, p)))
+    }
+    A <- new("matrix.csr", ra = c(rbind(-1/h, 1/h)), ja = as.integer(c(rbind(1:nh, 
+        2:(nh + 1)))), ia = as.integer(2 * (1:(nh + 1)) - 1), 
+        dimension = as.integer(c(nh, nh + 1)))
+    if (length(xun) == length(x)) {
+        F <- new("matrix.csr", ra = rep(1, nx), ja = as.integer(rank(x)), 
+            ia = 1:(nx + 1), dimension = as.integer(c(nx, nx)))
+    }
+    else {
+        F <- new("matrix.csr", ra = rep(1, nx), ja = as.integer(factor(x)), 
+            ia = 1:(nx + 1), dimension = as.integer(c(nx, length(xun))))
+    }
+    switch(constraint, 
+	   I = {R <- makeD(p); r <- rep(0, p-1)},
+	   D = {R <- -makeD(p); r <- rep(0, p-1)},
+	   N = {R <- NULL; r <- NULL})
+    list(x = list(x = xun), F = F[, -1], lambda = lambda, A = A[,
+        -1], Dorder = 0, R = R[,-1], r = r)
+}
 
 "qss1" <-
 function (x, constraint = "N", lambda = 1, dummies = dummies,
@@ -120,7 +163,7 @@ function (x, constraint = "N", lambda = 1, dummies = dummies,
 # Sparse Additive Quantile Smoothing Spline Models - Univariate Module
 #
 # This function returns a structure intended to make model.matrix for a univariate
-# nonparametric component of a model formula specified by a call to rq().  A sparse form
+# nonparametric component of a model formula specified by a call to rqss().  A sparse form
 # of the Frisch Newton algorithm is eventually called to compute the estimator.
 # Optional monotonicity and/or convexity/concavity constraints can be specified.  If
 # the formula consists of a single qss component then the estimator solves the
@@ -180,7 +223,6 @@ function (x, constraint = "N", lambda = 1, dummies = dummies,
         F <- new("matrix.csr", ra = rep(1, nx), ja = as.integer(factor(x)),
             ia = 1:(nx + 1), dimension = as.integer(c(nx, length(xun))))
 	}
-
    switch(constraint,
         V = {   R <- A;
                 r <- rep(0,nrow(R))
@@ -212,17 +254,30 @@ function (x, constraint = "N", lambda = 1, dummies = dummies,
 		},
 	N = { R=NULL; r=NULL}
 	)
-   list(x = list(x=xun), F=F[,-1], lambda = lambda, A=A[,-1], R=R[,-1], r=r)
+   list(x = list(x=xun), F=F[,-1], lambda = lambda, A=A[,-1], 
+	Dorder = 1, R=R[,-1], r=r)
 }
 "plot.qss1" <-
 function(x, rug = TRUE, jit = TRUE, add = FALSE, ...)
 {
-if(!add) plot(x[,1],x[,2],type = "n", ...)
-lines(x[,1],x[,2], ...)
-if(rug) {
-    if(jit)
-       rug(jitter(x[,1]))
-    else rug(x[,1])
+    if(!add) plot(x,type = "n", ...)
+	lines(x, ...)
+    if(rug) {
+	if(jit)
+	    rug(jitter(x[,1]))
+	else rug(x[,1])
+    }
+}
+"plot.qts1" <-
+function (x, rug = TRUE, jit = TRUE, add = FALSE, ...) 
+{
+    if (!add) 
+        plot(x, type = "S", ...)
+    lines(x, type = "S", ...)
+    if (rug) {
+        if (jit) 
+            rug(jitter(x[, 1]))
+        else rug(x[, 1])
     }
 }
 "plot.qss2" <-
@@ -356,62 +411,67 @@ function (x, rug = TRUE, jit = TRUE, bands = NULL, coverage = 0.95, add = FALSE,
     if(!length(select)) 
         select <- 1:m
     for (i in select) {
-        qss <- x$qss[[i]]$xyz
+	if(!is.null(x$qss[[i]]$Dorder)){ 
+	    if(x$qss[[i]]$Dorder == 0){
+		qts <- x$qss[[i]]$xyz
+		qts[, 2] <- x$coef[1] + qts[, 2]
+		plot.qts1(qts, add = add, ...)
+		}
+	    else { 
+		qss <- x$qss[[i]]$xyz
+		if (length(bands)) {
+		    if (is.na(x$coef["(Intercept)"])) 
+			stop("rqss confidence bands require an intercept parameter")
+		    B <- summary(x$qss[[i]], V[[i]], ...)
+		    cv <- B$cv
+		    B <- B$pred
+		    B$y <- B$y + x$coef["(Intercept)"]
+		    if(!length(bcol)) bcol <- c("grey85","grey65")
+		    for(k in 1:length(cv)){
+			if (add || k > 1) 
+			    if(shade){
+				polygon(c(B$x,rev(B$x)),
+				c(B$y - cv[k] * B$se,rev(B$y + cv[k] * B$se)),
+				col = bcol[k], border = FALSE)
+			    }
+			    else
+				matlines(B$x, cbind(B$y, B$y + cv[k] * cbind(-B$se, 
+				B$se)), lty = c(1, 2, 2), col = c("black", "blue", "blue"))
+			else {
+			    matplot(B$x, B$y + cv[k] * cbind(-B$se, B$se), 
+				xlab = paste(qssnames[i]), ylab = "Effect", type = "n", ...)
+			    if(shade){
+				polygon(c(B$x,rev(B$x)),
+				    c(B$y - cv[k] * B$se,rev(B$y + cv[k] * B$se)),
+				    col = bcol[k], border = FALSE)
+			    } 
+			    else{
+				lines(B$x, B$y + cv * B$se, lty = 2, ...)
+				lines(B$x, B$y - cv * B$se, lty = 2, ...)
+			    }
+			}
+			lines(B$x, B$y, ...)
+		    }
+		    band[[i]] <- list(x = B$x, blo = B$y - B$se %o% cv, bhi = B$y + B$se %o% cv)
+		    if (rug) {
+			if (jit) 
+			    rug(jitter(qss[, 1]))
+			else rug(qss[, 1])
+                    }
+		}
+		else {
+		    qss[, 2] <- x$coef[1] + qss[, 2]
+		    plot.qss1(qss, xlab = paste(qssnames[i]), ylab = "Effect", 
+		    rug = rug, jit = jit, add = add, ...)
+		}
+	    }
+        }
+	qss <- x$qss[[i]]$xyz
         if (ncol(qss) == 3) {
-            qss[, 3] <- x$coef[1] + qss[, 3]
-            plot.qss2(qss, ...)
+             qss[, 3] <- x$coef[1] + qss[, 3]
+             plot.qss2(qss, ...)
         }
-        else if (ncol(qss) == 2) {
-            if (length(bands)) {
-                if (is.na(x$coef["(Intercept)"])) 
-                  stop("rqss confidence bands require an intercept parameter")
-                B <- summary(x$qss[[i]], V[[i]], ...)
-                cv <- B$cv
-                B <- B$pred
-                B$y <- B$y + x$coef["(Intercept)"]
-                if(!length(bcol)) bcol <- c("grey85","grey65")
-		for(k in 1:length(cv)){
-                 if (add || k > 1) 
-                  if(shade){
-                     polygon(c(B$x,rev(B$x)),
-                        c(B$y - cv[k] * B$se,rev(B$y + cv[k] * B$se)),
-			col = bcol[k], border = FALSE)
-			}
-                  else
-                    matlines(B$x, cbind(B$y, B$y + cv[k] * cbind(-B$se, 
-                        B$se)), lty = c(1, 2, 2), col = c("black", "blue", "blue"))
-                else {
-                  matplot(B$x, B$y + cv[k] * cbind(-B$se, B$se), 
-                    xlab = paste(qssnames[i]), ylab = "Effect", 
-                    type = "n", ...)
-                  if(shade){
-                     polygon(c(B$x,rev(B$x)),
-                        c(B$y - cv[k] * B$se,rev(B$y + cv[k] * B$se)),
-			col = bcol[k], border = FALSE)
-			}
-                  else{
-                     lines(B$x, B$y + cv * B$se, lty = 2, ...)
-                     lines(B$x, B$y - cv * B$se, lty = 2, ...)
-		     }
-	           }
-                  lines(B$x, B$y, ...)
-                }
-                band[[1]] <- list(x = B$x, blo = B$y - B$se %o% cv, 
-                  bhi = B$y + B$se %o% cv)
-                if (rug) {
-                  if (jit) 
-                    rug(jitter(qss[, 1]))
-                  else rug(qss[, 1])
-                }
-            }
-            else {
-                qss[, 2] <- x$coef[1] + qss[, 2]
-                plot.qss1(qss, xlab = paste(qssnames[i]), ylab = "Effect", 
-                  rug = rug, jit = jit, add = add, ...)
-            }
-            title(titles[i])
-        }
-        else stop("invalid fitted qss object")
+	title(titles[i])
     }
     if (pages > 0) 
         par(oldpar)
@@ -501,7 +561,7 @@ function (object, newdata, interval = "none",  level = 0.95, ...)
     Names <- all.vars(parse(text = ff))
     if (any(!(Names %in% names(newdata)))) 
         stop("newdata doesn't include some model variables")
-    ff <- reformulate(ff)
+    #ff <- reformulate(ff)
     nd <- eval(model.frame(ff, data = newdata), parent.frame())
     qssterms <- attr(Terms, "specials")$qss
     if (length(qssterms)) {
@@ -638,13 +698,13 @@ predict.qss2 <- function (object, newdata, ...)
 fitted.rqss  <-  function(object, ...) (object$X %*% object$coef)[1:object$n]
 resid.rqss <- function(object, ...) object$resid[1:object$n]
 
-"rqss" <- function (formula, tau = 0.5, data = parent.frame(), weights, 
-    na.action, method = "sfn", lambda = NULL, contrasts = NULL, 
-    ztol = 1e-05,  control = sfn.control(), ...) 
+rqss <- function (formula, tau = 0.5, data = parent.frame(), weights, 
+    subset, na.action, method = "sfn", lambda = NULL, contrasts = NULL, 
+    ztol = 1e-05, control = sfn.control(), ...) 
 {
     call <- match.call()
     m <- match.call(expand.dots = FALSE)
-    temp <- c("", "formula", "data", "weights", "na.action")
+    temp <- c("", "formula", "data", "weights", "na.action", "subset")
     m <- m[match(temp, names(m), nomatch = 0)]
     m[[1]] <- as.name("model.frame")
     special <- "qss"
@@ -653,10 +713,10 @@ resid.rqss <- function(object, ...) object$resid[1:object$n]
     else terms(formula, special, data = data)
     qssterms <- attr(Terms, "specials")$qss
     dropx <- NULL
-    if(length(tau) > 1){
+    if (length(tau) > 1) {
         tau <- tau[1]
         warning("multiple taus not supported, using first element")
-        }
+    }
     if (length(qssterms)) {
         tmpc <- untangle.specials(Terms, "qss")
         ord <- attr(Terms, "order")[tmpc$terms]
@@ -666,33 +726,42 @@ resid.rqss <- function(object, ...) object$resid[1:object$n]
         if (length(dropx)) 
             Terms <- Terms[-dropx]
         attr(Terms, "specials") <- tmpc$vars
-        fnames <- function(x){
-           fy <- all.names(x[[2]])
-           if(fy[1] == "cbind") fy <- fy[-1]
-           fy
-           }
+        fnames <- function(x) {
+            fy <- all.names(x[[2]])
+            if (fy[1] == "cbind") 
+                fy <- fy[-1]
+            fy
+        }
         fqssnames <- unlist(lapply(parse(text = tmpc$vars), fnames))
         qssnames <- unlist(lapply(parse(text = tmpc$vars), function(x) deparse(x[[2]])))
     }
-    m$formula <- Terms
     ff <- delete.response(terms(formula(Terms)))
-    if(exists("fqssnames")){
-       ffqss <- paste(fqssnames, collapse = "+")
-       ff <- paste(deparse(formula(ff)), "+", ffqss)
+    if (exists("fqssnames")) {
+	mff <- m
+	mff$formula <- Terms
+	#mff <- delete.response(terms(formula(Terms)))
+        ffqss <- paste(fqssnames, collapse = "+")
+	#ff <- paste(deparse(formula(mff)), "+", ffqss)
+	mff$formula <- as.formula(paste(deparse(mff$formula), "+", ffqss))
+	ff <- delete.response(terms(formula(mff)))
+	ff <- deparse(formula(ff))
+	mff <- eval(mff, parent.frame())
     }
+    m$formula <- Terms
     m <- eval(m, parent.frame())
     weights <- model.extract(m, weights)
     process <- (tau < 0 || tau > 1)
     Y <- model.extract(m, "response")
-    if(requireNamespace("MatrixModels") && requireNamespace("Matrix")){
-         X <- MatrixModels::model.Matrix(Terms, m, contrasts.arg = contrasts, sparse = TRUE)
-         vnames <- dimnames(X)[[2]]
-         X <- as(X ,"matrix.csr")
-         }
-    else{
-         X <- model.matrix(Terms, m, contrasts)
-         vnames <- dimnames(X)[[2]]
-         }
+    if (requireNamespace("MatrixModels") && requireNamespace("Matrix")) {
+        X <- MatrixModels::model.Matrix(Terms, m, contrasts.arg = contrasts, 
+            sparse = TRUE)
+        vnames <- dimnames(X)[[2]]
+        X <- as(X, "matrix.csr")
+    }
+    else {
+        X <- model.matrix(Terms, m, contrasts)
+        vnames <- dimnames(X)[[2]]
+    }
     p <- ncol(X)
     pf <- environment(formula)
     nrL <- 0
@@ -713,8 +782,8 @@ resid.rqss <- function(object, ...) object$resid[1:object$n]
     }
     if (length(qssterms) > 0) {
         F <- as.matrix.csr(X)
-        qss <- lapply(tmpc$vars, function(u) eval(parse(text = u), 
-            data, enclos = pf))
+	pf <- environment(formula)
+        qss <- lapply(tmpc$vars, function(u) eval(parse(text = u), mff, pf))
         mqss <- length(qss)
         ncA <- rep(0, mqss + 1)
         nrA <- rep(0, mqss + 1)
@@ -754,31 +823,39 @@ resid.rqss <- function(object, ...) object$resid[1:object$n]
             R <- NULL
             r <- NULL
         }
-        if (method == "lasso") {
+        if (method == "lasso") 
             A <- rbind(cbind(L, as.matrix.csr(0, nrL, ncol(F) - ncL)), A)
-        }
+	if(length(weights)){
+	   F <- F * weights
+	   Y <- Y * weights
+	}
         X <- rbind(F, A)
         Y <- c(Y, rep(0, nrow(A)))
         rhs <- t(rbind((1 - tau) * F, 0.5 * A)) %*% rep(1, nrow(X))
         XpX <- t(X) %*% X
         nnzdmax <- XpX@ia[length(XpX@ia)] - 1
-	if(is.null(control[["nsubmax"]]))
-           control[["nsubmax"]] <- max(nnzdmax, floor(1000 + exp(-1.6) * nnzdmax^1.2))
-        if(is.null(control[["nnzlmax"]]))
-           control[["nnzlmax"]] <- floor(2e+05 - 2.8 * nnzdmax + 7e-04 * nnzdmax^2)
-        if(is.null(control[["tmpmax"]]))
-           control[["tmpmax"]] <- floor(1e+05 + exp(-12.1) * nnzdmax^2.35)
+        if (is.null(control[["nsubmax"]])) 
+            control[["nsubmax"]] <- max(nnzdmax, floor(1000 + 
+                exp(-1.6) * nnzdmax^1.2))
+        if (is.null(control[["nnzlmax"]])) 
+            control[["nnzlmax"]] <- floor(2e+05 - 2.8 * nnzdmax + 
+                7e-04 * nnzdmax^2)
+        if (is.null(control[["tmpmax"]])) 
+            control[["tmpmax"]] <- floor(1e+05 + exp(-12.1) * 
+                nnzdmax^2.35)
         fit <- if (length(r) > 0) 
             rqss.fit(X, Y, tau = tau, rhs = rhs, method = "sfnc", 
                 R = R, r = r, control = control, ...)
-        else rqss.fit(X, Y, tau = tau, rhs = rhs, method = "sfn", control = control, ...)
+        else rqss.fit(X, Y, tau = tau, rhs = rhs, method = "sfn", 
+            control = control, ...)
         for (i in 1:mqss) {
             ML <- p + 1 + ncA[i]
             MU <- p + ncA[i + 1]
             qss[[i]] <- list(xyz = cbind(qss[[i]]$x$x, qss[[i]]$x$y, 
-                c(0, fit$coef[ML:MU])), dummies = qss[[i]]$dummies)
+                c(0, fit$coef[ML:MU])), dummies = qss[[i]]$dummies, 
+			     Dorder = qss[[i]]$Dorder)
             if (ncol(qss[[i]]$xyz) == 2) 
-                class(qss[[i]]) <- "qss1"
+                class(qss[[i]]) <- ifelse(qss[[i]]$Dorder == 1,"qss1", "qts1")
             else class(qss[[i]]) <- "qss2"
         }
         names(qss) <- qssnames
@@ -787,42 +864,43 @@ resid.rqss <- function(object, ...) object$resid[1:object$n]
     else {
         X <- as.matrix.csr(X)
         nrA <- 0
-        if (method == "lasso") { #Pure lasso case
-            rhs <- t(rbind((1 - tau) * X, 0.5 * L)) %*% rep(1, nrow(X) + nrow(L))
+        if (method == "lasso") {
+            rhs <- t(rbind((1 - tau) * X, 0.5 * L)) %*% rep(1, 
+                nrow(X) + nrow(L))
             X <- rbind(X, L)
             Y <- c(Y, rep(0, nrL))
-            #nrA <- c(nrA, nrL) # Why was this here?
         }
-        else
-            rhs <- NULL
+        else rhs <- NULL
         if (length(weights)) {
             if (any(weights < 0)) 
                 stop("negative weights not allowed")
             X <- X * weights
             Y <- Y * weights
         }
-	XpX <- t(X) %*% X
+        XpX <- t(X) %*% X
         nnzdmax <- XpX@ia[length(XpX@ia)] - 1
-        if (is.null(control[["nsubmax"]]))
-            control[["nsubmax"]] <- max(nnzdmax, floor(1000 +
+        if (is.null(control[["nsubmax"]])) 
+            control[["nsubmax"]] <- max(nnzdmax, floor(1000 + 
                 exp(-1.6) * nnzdmax^1.2))
-        if (is.null(control[["nnzlmax"]]))
-            control[["nnzlmax"]] <- floor(2e+05 - 2.8 * nnzdmax +
+        if (is.null(control[["nnzlmax"]])) 
+            control[["nnzlmax"]] <- floor(2e+05 - 2.8 * nnzdmax + 
                 7e-04 * nnzdmax^2)
-        if (is.null(control[["tmpmax"]]))
-            control[["tmpmax"]] <- floor(1e+05 + exp(-12.1) *
+        if (is.null(control[["tmpmax"]])) 
+            control[["tmpmax"]] <- floor(1e+05 + exp(-12.1) * 
                 nnzdmax^2.35)
-        fit <- rqss.fit(X, Y, tau = tau,  rhs = rhs, control = control, 
-			method = method, ...)
+        fit <- rqss.fit(X, Y, tau = tau, rhs = rhs, control = control, 
+            method = method, ...)
         fit$nrA <- nrA
     }
     names(fit$coef) <- vnames
     n <- length(fit$resid) - nrL - nrA[length(nrA)]
+    if(length(weights)) fit$resid[1:n] <- fit$resid[1:n]/weights
     uhat <- fit$resid[1:n]
     Rho <- function(u, tau) sum(u * (tau - (u < 0)))
     fit$fidelity <- Rho(uhat, tau)
     fit$edf <- sum(abs(uhat) < ztol)
     fit$X <- X
+    fit$y <- Y
     fit$n <- n
     fit$nrL <- nrL
     fit$terms <- Terms
@@ -842,7 +920,17 @@ resid.rqss <- function(object, ...) object$resid[1:object$n]
     class(fit) <- "rqss"
     fit
 }
-
+"Munge" <- function(formula, ...) { 
+    # Recursive substitution {per Gabor Grothendieck}
+    if (length(formula) > 1) {
+	if (identical(formula[[2]], as.name(names(list(...))))) 
+	    formula <- eval(formula, list(...))
+	if (length(formula) > 1) 
+	    for (i in 1:length(formula)) 
+		formula[[i]] <- Recall(formula[[i]], ...)
+	}
+   formula
+}
 
 "summary.rqss" <- function(object, cov = FALSE, ztol = 1e-5, ...){
     resid <- object$resid
